import datetime
from typing import Dict


class Event(object):
    def __init__(self, data: Dict = {}):
        self.data = data
        self.created_at = datetime.datetime.utcnow()

    def __str__(self) -> str:
        return str(self.data)



from iamzero.config import Config
from iamzero.event import Event
from iamzero.client import Client
from iamzero.instrumentation import *
import os
import atexit
from typing import Dict, Optional

# we store the iamzero client as a global to avoid initialising
# multiple clients
_IAMZERO_CLIENT: Optional[Client] = None
_INITPID = None

WARNED_UNINITIALIZED = False


def init(token: str = None, url: str = None, debug: bool = None):
    global _IAMZERO_CLIENT
    global _INITPID

    pid = os.getpid()
    if _IAMZERO_CLIENT:
        if pid == _INITPID:
            _IAMZERO_CLIENT.log("iamzero is already initialised, skipping init")
            return
        else:
            _IAMZERO_CLIENT.log(
                f"iamzero already initialised, but process ID has changed (previously {_INITPID}, now {pid}"
            )
            _IAMZERO_CLIENT.close()

    config = Config(token=token, url=url, debug=debug)
    _IAMZERO_CLIENT = Client(config=config)
    _INITPID = pid


def get_client():
    return _IAMZERO_CLIENT


def send_event(data: Dict):
    event = Event(data=data)
    client = get_client()
    if client:
        client.send(event)


def fetch_identity(access_key=None, secret_key=None, token=None):
    client = get_client()
    if client:
        client.identity_fetcher.fetch_identity(
            access_key=access_key, secret_key=secret_key, token=token
        )


def get_responses_queue():
    client = get_client()
    if client:
        return client.responses()


def _flush():
    """
    Allows iamzero to be used in shorter Python scripts which exit immediately.
    We block the main thread until any pending messages have been flushed.
    """
    client = get_client()
    if client:
        client.close()


atexit.register(_flush)



from iamzero.logging import configure_root_logger
from iamzero.config import Config
from iamzero.identity import Identity, IdentityFetcher
from iamzero.event import Event

from iamzero.publisher import Publisher

logger = configure_root_logger(__name__)


class Client(object):
    """
    An iamzero client which is used to dispatch authentication errors to an iamzero collector
    """

    def __init__(
        self,
        config: Config,
        block_on_send=False,
        block_on_response=False,
        transmission_impl=None,
    ):
        self.config = config
        self.identity = Identity()
        self.identity_fetcher = IdentityFetcher(
            identity=self.identity, debug=self.config["debug"]
        )
        self.identity_fetcher.start()

        self.publisher = transmission_impl
        if self.publisher is None:
            self.publisher = Publisher(
                config=self.config,
                identity=self.identity,
                block_on_send=block_on_send,
                block_on_response=block_on_response,
            )

        self.publisher.start()
        self._responses = self.publisher.get_response_queue()
        self.block_on_response = block_on_response

        self.log(
            "initialized iamzero client: token=%s, url=%s",
            self.config["token"],
            self.config["url"],
        )
        if not self.config["token"]:
            self.log("token not set! set the token if you want to send data to iamzero")

    def log(self, msg, *args, **kwargs):
        if self.config["debug"]:
            logger.debug(msg, *args, **kwargs)

    def responses(self):
        """Returns a queue from which you can read a record of response info from
        each event sent. Responses will be dicts with the following keys:

        - `status_code` - the HTTP response from the api (eg. 200 or 503)
        - `duration` - how long it took to POST this event to the api, in ms
        - `metadata` - pass through the metadata you added on the initial event
        - `body` - the content returned by API (will be empty on success)
        - `error` - in an error condition, this is filled with the error message

        When the Client's `close` method is called, a None will be inserted on
        the queue, indicating that no further responses will be written.
        """
        return self._responses

    def send(self, event: Event):
        """
        Enqueues the given event to be sent
        """
        if self.publisher is None:
            self.log(
                "tried to send on a closed or uninitialized iamzero client,"
                " event = %s",
                event,
            )
            return

        self.log("send enqueuing event, event = %s", event)
        self.publisher.send(event)

    def close(self):
        """Wait for in-flight events to be transmitted then shut down cleanly.
        Optional (will be called automatically at exit) unless your
        application is consuming from the responses queue and needs to know
        when all responses have been received."""

        if self.publisher:
            self.publisher.close()

        if self.identity_fetcher:
            self.identity_fetcher.close()

        # set to None so that any future sends throw errors
        self.publisher = None
        self.identity_fetcher = None

    def flush(self):
        """Closes and restarts the transmission, sending all events. Use this
        if you want to perform a blocking send of all events in your
        application.
        """
        if self.publisher and isinstance(self.publisher, Publisher):
            self.publisher.close()
            self.publisher.start()



import logging

ROOT_LOGGER_NAME = "iamzero"
LOG_FORMAT = (
    "[%(levelname)s][%(asctime)s #%(process)d.%(threadName)s]"
    " %(name)s:%(lineno)s \t%(message)s"
)


def configure_root_logger(name, log_level=logging.DEBUG):
    """
    configure iamzero logger
    """
    logger = logging.getLogger(name)

    # Don't propagate messages to upper loggers
    logger.propagate = False
    logger.handlers = []

    formatter = logging.Formatter(LOG_FORMAT)

    # Configure the stderr handler
    stderr_handler = logging.StreamHandler()
    stderr_handler.setFormatter(formatter)
    logger.addHandler(stderr_handler)

    try:
        logger.setLevel(log_level)
    except ValueError:
        logger.error("Unknown log_level %r, default log level is CRITICAL", log_level)
        logger.setLevel(logging.CRITICAL)

    return logger



import iamzero
from wrapt import wrap_function_wrapper


def wrapped_api_call(wrapped, instance, args, kwargs):
    try:
        result = wrapped(*args, **kwargs)
        return result
    except Exception as e:
        # add error to queue to be dispatched
        data = {
            "service": instance._service_model.service_name,
            "region": instance.meta.region_name,
            "operation": args[0],
            "parameters": args[1],
            "exceptionMessage": e.response.get("Error", {}).get("Message"),
            "exceptionCode": e.response.get("Error", {}).get("Code"),
        }
        iamzero.send_event(data=data)
        raise


def wrapped_client_creator(wrapped, instance, args, kwargs):
    # By default, botocore does not store the role ARN of the session.
    # We need this so that iamzero can tell **which** role or user
    # had the permissions error, so we know what to create
    # least-privilege recommendations for.
    #
    # We instrument the client creator method to capture the AWS credentials.
    # These are used to make a call to the AWS STS get-caller-identity endpoint
    # in a background thread, to identify the role ARN of the session.
    #
    # These credentials are only used to call AWS resources and aren't dispatched
    # to the iamzero collector.
    credentials = kwargs["credentials"]
    access_key = credentials.access_key
    secret_key = credentials.secret_key
    token = credentials.token

    service_name = kwargs["service_name"]

    # don't fetch identity for the sts service, as iamzero uses a botocore
    # client under the hood to fetch the identity. This would cause an infinite
    # instrumentation loop.
    if service_name != "sts":
        iamzero.fetch_identity(
            access_key=access_key, secret_key=secret_key, token=token
        )

    return wrapped(*args, **kwargs)


wrap_function_wrapper("botocore.client", "BaseClient._make_api_call", wrapped_api_call)

wrap_function_wrapper(
    "botocore.client", "ClientCreator.create_client", wrapped_client_creator
)



import logging
import os

from configparser import Error as ConfigError, RawConfigParser
from typing import Any, Mapping, Optional


LOGGER = logging.getLogger(__name__)

CONFIG_DEFAULT_VALUES = {
    "url": "https://app.iamzero.dev",
    "debug": False,
    "token": None,
    "max_batch_size": 100,
    "send_frequency": 0.25,
    "user_agent_addition": "",
}


HOME_FILE_PATH = os.path.expanduser("~/.iamzero.ini")


class Config(object):
    """
    Loads config with the following priorities:

    1. keyword arguments that Config was called with (for example - `Config(url="test")` will always set the url variable to test)
    2. env variables (IAMZERO_ + the name of the config variable - eg IAMZERO_URL)
    2. config file (by default ~/.iamzero.ini, or whatever the value of IAMZERO_CONFIG_FILE env var is)
    3. Default values as per CONFIG_DEFAULT_VALUE
    """

    FILE_ENV_VAR = "IAMZERO_CONFIG_FILE"
    HOME_FILE_PATH = HOME_FILE_PATH

    def __init__(self, **kwargs):
        self._init_args = kwargs

        self.config = {}  # type: (Mapping[str, Any])

        self.default_values = CONFIG_DEFAULT_VALUES

        self.loaders = [
            self.load_from_default_values,
            self.load_from_file,
            self.load_from_env,
            self.load_from_init_args,
        ]

        self.config_path = None
        self.load()

    def load(self):
        """Call each loader and update the config variable at the end"""
        base_config = {}

        for loader in self.loaders:
            loaded = loader()

            if loaded:
                base_config.update(loaded)

        self.config = base_config

    def load_from_default_values(self):
        """Returns default values"""
        return self.default_values

    def load_from_init_args(self):
        """Replaces config with the kwargs that __init__ was called with"""
        config_dict = {}
        for key, val in self._init_args.items():
            if val is not None:
                config_dict[key] = val

        return config_dict

    def load_from_file(self):
        """
        Load from config file
        """

        file_path = (
            self.config_path
            or self._file_path_from_env()
            or self._file_path_from_home()
        )

        if not file_path:
            return {}

        config = RawConfigParser()

        try:
            config.read(file_path)

            config_dict = {}
            for option in config.options("iamzero"):
                config_dict[option] = self._coerce_value(
                    option, config.get("iamzero", option)
                )

            return config_dict
        except ConfigError:
            LOGGER.debug("Error parsing config file %s", file_path)
            return {}

    def load_from_env(self):
        """Load configuration from os environment variables, variables
        must be prefixed with IAMZERO_ to be detected.
        """
        env_config = {}
        for env_var, value in os.environ.items():
            if env_var.startswith("IAMZERO_"):
                key = env_var[8:].upper()
                env_config[key] = self._coerce_value(key, value)
        return env_config

    def _coerce_value(self, name, value):  # type: (str, Any) -> Any
        default_value = self.default_values.get(name)
        # best effort conversion to boolean
        if isinstance(default_value, bool):
            value = value.lower().strip() in ("1", "true", "yes", "y")
        elif isinstance(default_value, (int, float)):
            try:
                value = type(default_value)(value)
            except ValueError:
                LOGGER.error(
                    "Invalid config value for %s, using default value", name.lower()
                )
                value = default_value
        return value

    def _file_path_from_env(self):
        """Return file path if os environement was set and file exists"""
        path = os.getenv(self.FILE_ENV_VAR, default=None)

        if path and os.path.isfile(path):
            return path

    def _file_path_from_home(self):
        """Return file path if file exists in home directory"""
        if self.HOME_FILE_PATH and os.path.isfile(self.HOME_FILE_PATH):
            return self.HOME_FILE_PATH

    def __getitem__(self, name):  # type: (str) -> Any
        return self.config[name]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iamzero', 'iamzero.instrumentation']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.1,<3.0.0', 'statsd>=3.3.0,<4.0.0', 'wrapt>=1.12.1,<2.0.0']

setup_kwargs = {
    'name': 'iamzero',
    'version': '0.0.4',
    'description': 'iam-zero least-privilege instrumentation client for Python',
    'long_description': '<p align="center"><img src="https://iamzero.dev/img/logo-boxed.svg" height="80" /></p>\n\n<h1 align="center">iamzero-python</h1>\n<p align="center">Python client library for <a href="https://iamzero.dev">IAM Zero</a></p>\n\n<p align="center">\n<a align="center" href="https://twitter.com/intent/tweet?url=https%3A%2F%2Fiamzero.dev&text=IAM%20Zero%20simplifies%20cloud%20permissions%20for%20development%20teams"><img src="https://img.shields.io/twitter/url/https/github.com/tterb/hyde.svg?style=social" alt="tweet" /></a>\n<a href="https://join.slack.com/t/commonfatecommunity/shared_invite/zt-q4m96ypu-_gYlRWD3k5rIsaSsqP7QMg"><img src="https://img.shields.io/badge/slack-iamzero-1F72FE.svg?logo=slack" alt="slack" /></a>\n</p>\n\n<p align="center">\n    <a href="https://iamzero.dev/docs/getting-started">🚀 Get Started</a> |\n    <a href="https://iamzero.dev/docs/support">📖 Support</a>\n</p>\n\nThis package makes it easy to create least-privilege IAM policies for your Python scripts or applications.\n\n## What is IAM Zero?\n\nIAM Zero detects identity and access management issues and automatically suggests least-privilege policies. It does this by capturing errors in applications you build or commands that you run which use. By detecting the error and matching it against our Access Advisory lists IAM Zero can instantly provide a least-privilege policy recommendation, customised to your cloud environment.\n\nIAM Zero currently works for AWS but our roadmap includes other cloud platforms like GCP, Azure, and Kubernetes.\n\n## Documentation\n\nGet started by [reading our documentation](https://iamzero.dev/docs/getting-started).\n\n## Contributing\n\nSee [CONTRIBUTING.md](./CONTRIBUTING.md) for information on how to contribute. We welcome all contributors - [join our Slack](https://join.slack.com/t/commonfatecommunity/shared_invite/zt-q4m96ypu-_gYlRWD3k5rIsaSsqP7QMg) to discuss the project!\n\n## Security\n\nYou can view our full security documentation on the [IAM Zero website](https://iamzero.dev/docs/security).\n',
    'author': 'Common Fate',
    'author_email': 'hello@commonfate.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://iamzero.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



from iamzero.logging import configure_root_logger
import threading
import queue
from botocore.session import get_session

logger = configure_root_logger(__name__)


class Identity(object):
    """
    Used to store AWS identity
    """

    def __init__(self) -> None:
        self.mutex = threading.RLock()
        self.initialized = threading.Condition(self.mutex)
        self.user = None
        self.role = None
        self.account = None

    def set(self, user: str = None, role: str = None, account: str = None):
        with self.initialized:
            self.user = user
            self.role = role
            self.account = account
            self.initialized.notify_all()


class IdentityRequest(object):
    """
    Stores AWS credentials used to make a request to fetch identity
    using AWS STS get-caller-identity
    """

    def __init__(self, access_key: str, secret_key: str, token: str):
        self.access_key = access_key
        self.secret_key = secret_key
        self.token = token


class IdentityFetcher(object):
    """
    Manages a background thread which fetches AWS identity
    using AWS STS get-caller-identity
    """

    def __init__(self, identity=None, debug=False) -> None:
        self.pending: queue.Queue[IdentityRequest] = queue.Queue(5)
        self.debug = debug
        self.identity: Identity = identity

        if self.identity is None:
            raise Exception("Identity must be provided")

    def log(self, msg, *args, **kwargs):
        if self.debug:
            logger.debug(msg, *args, **kwargs)

    def fetch_identity(self, access_key=None, secret_key=None, token=None):
        if access_key is None or secret_key is None or token is None:
            self.log("access_key, secret_key, and token must all be provided")
            return
        self.log("requesting identity")
        self.pending.put(
            IdentityRequest(access_key=access_key, secret_key=secret_key, token=token)
        )

    def start(self):
        self._sending_thread = threading.Thread(target=self._sender)
        self._sending_thread.daemon = True
        self._sending_thread.start()
        self.log("started identity fetcher thread")

    def _sender(self):
        while True:
            ev = self.pending.get()
            if ev is None:
                return
            session = get_session()
            sts = session.create_client(
                "sts",
                aws_access_key_id=ev.access_key,
                aws_secret_access_key=ev.secret_key,
                aws_session_token=ev.token,
            )
            identity = sts.get_caller_identity()
            self.identity.set(
                user=identity["UserId"],
                role=identity["Arn"],
                account=identity["Account"],
            )

    def close(self):
        """call close to send all in-flight requests and shut down the
        senders nicely. Times out after max 20 seconds per sending thread
        plus 10 seconds for the response queue"""
        try:
            self.pending.put(None, True, 10)
        except queue.Full:
            pass
        self._sending_thread.join()



from iamzero.logging import configure_root_logger
from iamzero.config import Config
from iamzero.identity import Identity
from typing import List
from iamzero.event import Event

import queue
from urllib.parse import urljoin
import gzip
import io
import json
import threading
import requests
import statsd
import time

from iamzero.version import VERSION

logger = configure_root_logger(__name__)


class Publisher:
    def __init__(
        self,
        config: Config,
        block_on_send=False,
        block_on_response=False,
        identity=None,
        gzip_enabled=False,
        gzip_compression_level=1,
        proxies={},
    ):
        self.config = config
        self.token = config["token"]
        self.block_on_send = block_on_send
        self.block_on_response = block_on_response
        self.max_batch_size = config["max_batch_size"]
        self.send_frequency = config["send_frequency"]
        self.gzip_compression_level = gzip_compression_level
        self.gzip_enabled = gzip_enabled
        self.url = config["url"]
        self.identity: Identity = identity

        if self.identity is None:
            raise Exception("Identity must be provided")

        user_agent = "iamzero-python/" + VERSION
        if self.config["user_agent_addition"]:
            user_agent += " " + self.config["user_agent_addition"]

        session = requests.Session()
        session.headers.update({"User-Agent": user_agent})
        if self.gzip_enabled:
            session.headers.update({"Content-Encoding": "gzip"})
        if proxies:
            session.proxies.update(proxies)
        self.session = session

        # pending events queue
        self.pending: queue.Queue[Event] = queue.Queue(maxsize=1000)
        # API responses queue
        self.responses = queue.Queue(maxsize=2000)

        self._sending_thread = None
        self.sd = statsd.StatsClient(prefix="iamzero")

        self.debug = self.config["debug"]

    def log(self, msg, *args, **kwargs):
        if self.debug:
            logger.debug(msg, *args, **kwargs)

    def start(self):
        self._sending_thread = threading.Thread(target=self._sender)
        self._sending_thread.daemon = True
        self._sending_thread.start()
        self.log("started sending thread")

    def send(self, ev: Event):
        """send accepts an event and queues it to be sent"""
        self.sd.gauge("queue_length", self.pending.qsize())
        try:
            if self.block_on_send:
                self.pending.put(ev)
            else:
                self.pending.put_nowait(ev)
            self.sd.incr("messages_queued")
        except queue.Full:
            response = {
                "status_code": 0,
                "duration": 0,
                "metadata": ev.metadata,
                "body": "",
                "error": "event dropped; queue overflow",
            }
            if self.block_on_response:
                self.responses.put(response)
            else:
                try:
                    self.responses.put_nowait(response)
                except queue.Full:
                    # if the response queue is full when trying to add an event
                    # queue is full response, just skip it.
                    pass
            self.sd.incr("queue_overflow")

    def _sender(self):
        """_sender is the control loop that pulls events off the `self.pending`
        queue and submits batches for actual sending."""
        events: List[Event] = []
        last_flush = time.time()

        # wait until identity is established before dispatching
        # events, so that the role can be included in events
        #
        # TODO: likely need to handle the case where identity is
        # not loaded properly rather than waiting forever and letting
        # the pending queue fill up.
        # TODO: how does this work with multiple identities if assume role
        # functionality is used?
        with self.identity.initialized:
            self.identity.initialized.wait()

        while True:
            try:
                ev = self.pending.get(timeout=self.send_frequency)
                if ev is None:
                    # signals shutdown
                    self._flush(events)
                    return
                events.append(ev)
                if (
                    len(events) > self.max_batch_size
                    or time.time() - last_flush > self.send_frequency
                ):
                    self._flush(events)
                    events = []
                    last_flush = time.time()
            except queue.Empty:
                self._flush(events)
                events = []
                last_flush = time.time()

    def _flush(self, events: List[Event]):
        if not events:
            return
        self._send_batch(events)

    def _send_batch(self, events: List[Event]):
        """Makes a single batch API request with the given list of events. The
        `destination` argument contains the write key, API host and dataset
        name used to build the request."""
        start = time.time()
        status_code = 0
        try:
            url = urljoin(self.url, "api/v1/events/")
            payload = []
            for ev in events:
                event_time = ev.created_at.isoformat()
                if ev.created_at.tzinfo is None:
                    event_time += "Z"
                payload.append(
                    {
                        "time": event_time,
                        "data": ev.data,
                        "identity": {
                            "user": self.identity.user,
                            "role": self.identity.role,
                            "account": self.identity.account,
                        },
                    }
                )

            data = json.dumps(payload)
            if self.gzip_enabled:
                # The gzip lib works with file-like objects so we use a buffered byte stream
                stream = io.BytesIO()
                compressor = gzip.GzipFile(
                    fileobj=stream, mode="wb", compresslevel=self.gzip_compression_level
                )
                compressor.write(data.encode())
                compressor.close()
                data = stream.getvalue()
                stream.close()
            self.log("firing batch, size = %d", len(payload))
            resp = self.session.post(
                url,
                headers={
                    "x-iamzero-token": self.token,
                    "Content-Type": "application/json",
                },
                data=data,
                timeout=10.0,
            )
            status_code = resp.status_code
            resp.raise_for_status()
            # log response to the responses queue
            self._enqueue_response(status_code, resp.json(), None, start, ev)

        except Exception as e:
            # Catch all exceptions and hand them to the responses queue.
            self._enqueue_errors(status_code, e, start, events)

    def _enqueue_errors(self, status_code, error, start, events):
        for ev in events:
            self.sd.incr("send_errors")
            self._enqueue_response(status_code, "", error, start, ev)

    def _enqueue_response(self, status_code, body, error, start, metadata):
        resp = {
            "status_code": status_code,
            "body": body,
            "error": error,
            "duration": (time.time() - start) * 1000,
            "metadata": metadata,
        }
        self.log("enqueuing response = %s", resp)
        if self.block_on_response:
            self.responses.put(resp)
        else:
            try:
                self.responses.put_nowait(resp)
            except queue.Full:
                pass

    def close(self):
        """call close to send all in-flight requests and shut down the
        senders nicely. Times out after max 20 seconds per sending thread
        plus 10 seconds for the response queue"""
        try:
            self.pending.put(None, True, 10)
        except queue.Full:
            pass
        self._sending_thread.join()
        # signal to the responses queue that nothing more is coming.
        try:
            self.responses.put(None, True, 10)
        except queue.Full:
            pass

    def get_response_queue(self):
        """return the responses queue on to which will be sent the response
        objects from each event send"""
        return self.responses
